//_____________________________________________________________________________
//
// Copyright (C) 2019                   Mobsya                   CH-1020 Renens
//_____________________________________________________________________________
//
// PROJECT   Thymio-III
//_____________________________________________________________________________
//
//! \file    mp3.c
//! \brief   This module provides the useful functions to use the MP3
//!
//! \author  Vincent Gonet
//!
//! \version $Id: mp3.c 18076 2017-04-20 12:28:12Z v.gonet $
//_____________________________________________________________________________

//-----------------------------------------------------------------------------
// Include Section
//-----------------------------------------------------------------------------

#include <string.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

#include "esp_log.h"

#include "audio_element.h"
#include "audio_pipeline.h"
#include "audio_event_iface.h"
#include "audio_mem.h"
#include "audio_common.h"
#include "i2s_stream.h"
#include "mp3_decoder.h"

#include "mp3.h"

//-----------------------------------------------------------------------------
// Constants/Macros Definitions
//-----------------------------------------------------------------------------

#define I2S_STREAM_INTERNAL_DAC_CFG_CUSTOM() {                                      \
    .type = AUDIO_STREAM_WRITER,                                                    \
    .task_prio = I2S_STREAM_TASK_PRIO,                                              \
    .task_core = I2S_STREAM_TASK_CORE,                                              \
    .task_stack = I2S_STREAM_TASK_STACK,                                            \
    .out_rb_size = I2S_STREAM_RINGBUFFER_SIZE,                                      \
    .i2s_config = {                                                                 \
        .mode = I2S_MODE_MASTER | I2S_MODE_DAC_BUILT_IN | I2S_MODE_TX,              \
        .sample_rate = 44100,                                                       \
        .bits_per_sample = 16,                                                      \
        .channel_format = I2S_CHANNEL_FMT_ALL_RIGHT,                                \
        .communication_format = I2S_COMM_FORMAT_I2S_MSB,                            \
        .dma_buf_count = 3,                                                         \
        .dma_buf_len = 300,                                                         \
        .intr_alloc_flags = 0,                                                      \
    },                                                                              \
    .i2s_port = 0,                                                                  \
}

//-----------------------------------------------------------------------------
// Types Definitions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Exported Global Data
//-----------------------------------------------------------------------------

extern const uint8_t adf_music_mp3_start[] asm("_binary_adf_music_mp3_start");
extern const uint8_t adf_music_mp3_end[] asm("_binary_adf_music_mp3_end");

//-----------------------------------------------------------------------------
// Private Data
//-----------------------------------------------------------------------------

static const char* Tag = "mp3";

audio_pipeline_handle_t pipeline;
audio_element_handle_t i2s_stream_writer;
audio_element_handle_t mp3_decoder;

//-----------------------------------------------------------------------------
// Private Functions Prototypes
//-----------------------------------------------------------------------------

int mp3_music_read_cb(audio_element_handle_t el, char* buf, int len, TickType_t wait_time, void* ctx);

//-----------------------------------------------------------------------------
// Inline Code Definition
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Functions Implementation
//-----------------------------------------------------------------------------

void MP3_Init(void)
{
  esp_log_level_set("*", ESP_LOG_WARN);
  esp_log_level_set(Tag, ESP_LOG_INFO);

  ESP_LOGI(Tag, "[ 1 ] Create audio pipeline, add all elements to pipeline, and subscribe pipeline event");
  audio_pipeline_cfg_t pipeline_cfg = DEFAULT_AUDIO_PIPELINE_CONFIG();
  pipeline = audio_pipeline_init(&pipeline_cfg);
  mem_assert(pipeline);

  ESP_LOGI(Tag, "[1.1] Create mp3 decoder to decode mp3 file and set custom read callback");
  mp3_decoder_cfg_t mp3_cfg = DEFAULT_MP3_DECODER_CONFIG();
  mp3_decoder = mp3_decoder_init(&mp3_cfg);
  audio_element_set_read_cb(mp3_decoder, mp3_music_read_cb, NULL);

  ESP_LOGI(Tag, "[1.2] Create i2s stream to write data to ESP32 internal DAC");
  i2s_stream_cfg_t i2s_cfg = I2S_STREAM_INTERNAL_DAC_CFG_CUSTOM();
  i2s_cfg.type = AUDIO_STREAM_WRITER;
  i2s_stream_writer = i2s_stream_init(&i2s_cfg);

  ESP_LOGI(Tag, "[1.3] Register all elements to audio pipeline");
  audio_pipeline_register(pipeline, mp3_decoder, "mp3");
  audio_pipeline_register(pipeline, i2s_stream_writer, "i2s");

  ESP_LOGI(Tag, "[1.4] Link it together [mp3_music_read_cb]-->mp3_decoder-->i2s_stream-->[ESP32 DAC]");
  audio_pipeline_link(pipeline, (const char* []) {"mp3", "i2s"}, 2);
}

//_____________________________________________________________________________

void MP3_Task(void)
{
  ESP_LOGI(Tag, "[ 2 ] Setup event listener");
  audio_event_iface_cfg_t evt_cfg = AUDIO_EVENT_IFACE_DEFAULT_CFG();
  audio_event_iface_handle_t evt = audio_event_iface_init(&evt_cfg);

  ESP_LOGI(Tag, "[2.1] Listening event from all elements of pipeline");
  audio_pipeline_set_listener(pipeline, evt);

  ESP_LOGI(Tag, "[ 3 ] Start audio_pipeline");
  audio_pipeline_run(pipeline);

  while (1)
  {
    audio_event_iface_msg_t msg;
    esp_err_t ret = audio_event_iface_listen(evt, &msg, portMAX_DELAY);

    if (ret != ESP_OK)
    {
      ESP_LOGE(Tag, "[ * ] Event interface error : %d", ret);
      continue;
    }

    if (msg.source_type == AUDIO_ELEMENT_TYPE_ELEMENT && msg.source == (void*) mp3_decoder &&
        msg.cmd == AEL_MSG_CMD_REPORT_MUSIC_INFO)
    {
      audio_element_info_t music_info = {0};
      audio_element_getinfo(mp3_decoder, &music_info);

      ESP_LOGI(Tag, "[ * ] Receive music info from mp3 decoder, sample_rates=%d, bits=%d, ch=%d",
               music_info.sample_rates, music_info.bits, music_info.channels);

      audio_element_setinfo(i2s_stream_writer, &music_info);
      i2s_stream_set_clk(i2s_stream_writer, music_info.sample_rates, music_info.bits, music_info.channels);
      continue;
    }

    /* Stop when the last pipeline element (i2s_stream_writer in this case) receives stop event */
    if (msg.source_type == AUDIO_ELEMENT_TYPE_ELEMENT && msg.source == (void*) i2s_stream_writer  &&
        msg.cmd == AEL_MSG_CMD_REPORT_STATUS && (int) msg.data == AEL_STATUS_STATE_STOPPED)
    {
      break;
    }
  }

  ESP_LOGI(Tag, "[ 4 ] Stop audio_pipeline");
  audio_pipeline_terminate(pipeline);

  audio_pipeline_unregister(pipeline, mp3_decoder);
  audio_pipeline_unregister(pipeline, i2s_stream_writer);

  /* Terminate the pipeline before removing the listener */
  audio_pipeline_remove_listener(pipeline);

  /* Make sure audio_pipeline_remove_listener is called before destroying event_iface */
  audio_event_iface_destroy(evt);

  /* Release all resources */
  audio_pipeline_deinit(pipeline);
  audio_element_deinit(i2s_stream_writer);
  audio_element_deinit(mp3_decoder);
  vTaskDelete(NULL);
}

//_____________________________________________________________________________

int mp3_music_read_cb(audio_element_handle_t el, char* buf, int len, TickType_t wait_time, void* ctx)
{
  static int mp3_pos;
  int read_size = adf_music_mp3_end - adf_music_mp3_start - mp3_pos;
  if (read_size == 0)
  {
    return AEL_IO_DONE;
  }
  else if (len < read_size)
  {
    read_size = len;
  }
  memcpy(buf, adf_music_mp3_start + mp3_pos, read_size);
  mp3_pos += read_size;
  return read_size;
}
